/*
 *	Title:		o3find
 *	Version:	0.70
 *	Created by:	Paolo Copello
 *	Home page:	http://web.tiscali.it/fanelia
 *	
 *	Copyright (C) 2003 Paolo Copello
 *	
 *	******************************************
 *	This file is part of o3find
 *	******************************************
 *	
 *	o3find is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	any later version.
 *	
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *	
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *	
 */





// suppress "truncated debug info" warnings (VC++ specific)
#ifdef _MSC_VER
#pragma warning (disable: 4786)
#pragma warning (disable: 4788)
#endif // _MSC_VER


#include "xmlmemfile.h"


namespace Fanel
{

/** Read the next "valid" chunk from the extracted XML document.
	A chunk is valid if its lenght is not zero and if it is not 
	a space or newline.
	The function ignores XML tags and comments. */
const char* OoXmlMemFile::ReadNextXmlChunk(int& nLenght)
{
	const char* pCurrentChunk = NULL;

	enum ReadState { RS_INPROGRESS, RS_NEWLINE, RS_TAG, RS_EOF };

	bool bValidChunk = false;

	do
	{
		if (m_nPos > m_nDataSize)
			return NULL;

		ReadState read_state = RS_INPROGRESS;
		size_t nStart = m_nPos;

		char* pBufAsChar = (char*)m_lpBuf;

	//	bFoundInChunk = false;
		while(read_state == RS_INPROGRESS)
		{
			char ch = *(pBufAsChar+m_nPos);
			m_nPos++;
			
			if (ch == '\n' || ch == '\r')
				read_state = RS_NEWLINE;
			else if (ch == '<')
				read_state = RS_TAG;
			else if (m_nPos > m_nDataSize)
				read_state = RS_EOF;
		}

		pCurrentChunk = pBufAsChar+nStart;
		nLenght = (pBufAsChar+m_nPos-1) - pCurrentChunk ; // TODO verify if this is correct

		if(nLenght>1 || (  nLenght==1 && *pCurrentChunk != ' ' && *pCurrentChunk != '\n'
			&& *pCurrentChunk != '\r') ) 
			bValidChunk = true;

		if(read_state == RS_TAG)
		{
			// loop until the > symbol is found
			bool bEnd = false;

			// ignore XML comments
			if( strncmp( pBufAsChar+m_nPos , "!--", 3) == 0)
			{
				m_nPos += 3;

				while(!bEnd)
				{
					int iComp = strncmp( pBufAsChar+m_nPos , "-->", 3);
					m_nPos++;
					if (iComp == 0 || m_nPos > m_nDataSize)
						bEnd = true;
				}
			}
			else
			{
				while(!bEnd)
				{
					char ch = *(pBufAsChar+m_nPos);
					m_nPos++;
					if (ch == '>' || m_nPos > m_nDataSize)
						bEnd = true;
				}
			}
		}
		else if (read_state == RS_NEWLINE)
		{
			// loop while adjacent newlines are found
			bool bEnd = false;
			while(!bEnd)
			{
				char ch = *(pBufAsChar+m_nPos);
				if (ch != '\n' || ch != '\r' || m_nPos > m_nDataSize)
					bEnd = true;
				else
					m_nPos++;
			}
		}
	} while(!bValidChunk);

	return pCurrentChunk;
}



}

